/** -------------------------------------------------------------------
  *      maintainer: Marcus Wolschon <Suran@gmx.net>
  * -------------------------------------------------------------------
  * projects:
  *	- Jaugment II
  *       part: Java3D-UI
  *	   used in release: current      used in version: current
  * -------------------------------------------------------------------
  * Copyright:   Copyright (c) 2000 by Marcus Wolschon <Suran@gmx.net>
  * -------------------------------------------------------------------
  * changes:
  * 0.0.1 (2000-03-31, Marcus Wolschon<Suran@gmx.net>)
  *       - created
  * -------------------------------------------------------------------
  * TODO:
  *       - implement more actions
  *       - implement functions for editing automata
  *       - make it cofigurable
  *       - let it simulate the former Window2DExamineViewerBehavior
  */
package jaugment2.UI.java3dUI.runeMagic;

  //JDK
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Stack;
import java.lang.reflect.Method;
  //Java3D
import javax.media.j3d.*;
import javax.media.j3d.WakeupOr;
import javax.media.j3d.WakeupOnCollisionEntry;
   //jAugment
import jaugment2.helpers.Logger;

/**
 * Title:        JAugment II - Java3d-UI
 * Description:  Jini-Based system of wearable-optimized services
 * Copyright:    Copyright (c) 1999
 * Company:
 * @author Marcus Wolschon Suran@gmx.net
 * @version
 */


/**
 * This class represents a behavior, that is not fixed
 * in the way it processes stimuli but executes a
 * finite automata where state-transitions are executed
 * on defined stimuli and lists of actions with calculated
 * arguments are executed uppon transitions.
 */
public class AutomataBehavior extends Behavior
{
 private static final boolean debutThis = false;

  protected void _wakeupOn(WakeupCondition p0)
  {
   if(debutThis)Logger.debug.println("DEBUG [AutomataBehavior] _wakeupOn(p0="+p0+")");
   wakeupOn(p0);
  }

  /**
   * Class representing a finite automata
   */
  public static class Automata
  {
   public Automata(State states[], State first)
   {
    this.firstState = first;
    this.states = new LinkedList();
    this.states.add(first);
    for(int i=0; i<states.length; i++)
      if(states[i] != first)
        this.states.add(states[i]);

    currentState = firstState;
    this.stack = new Stack();
   }

   public void addState(State s)
   {
    states.add(s);
   }

   public void setFirstState(State s, AutomataBehavior This)
   {
    if(!states.contains(s))
      states.add(s);
    if(currentState ==
     firstState)
      {
       currentState = s;
       firstState = s;
       activate(This);
      }
    firstState = s;
   }

   public void activate(AutomataBehavior This)
   {
    This._wakeupOn(currentState.getActiveCondition());
   };

   private LinkedList states;
   private State firstState;
   private State currentState;

   /**
    * actions can add/remove objects from/to the stack
    * for parameter-values
    */
   private Stack stack;


   public void processStimulus( Enumeration criteria, AutomataBehavior This)
   {
    State s = currentState.processStimulus(criteria, stack);
    if(s != currentState)
      {
       currentState = s;
      }
    This._wakeupOn(currentState.getActiveCondition());
   }

  }

  /**
   * Class representing a state of a
   * finite automata including the possible
   * transitions from here
   */
  public static class State
  {
   public State(Transition transitions[])
   {
    this.transitions = transitions;
    WakeupCriterion c[] = new WakeupCriterion[transitions.length];
    for(int i=0; i<transitions.length; i++)
      c[i] = transitions[i].getWakeupCriterion();
    this.activeCriterion = new WakeupOr(c);

    if(debutThis)Logger.debug.println("DEBUG [State] init #criteria="+c.length+"");
   }

   public void addTransition(Transition trans)
   {
    Transition transitions2[] = new Transition[transitions.length+1];
    for(int i=0; i<transitions.length; i++)
      transitions2[i] = transitions[i];
    transitions2[transitions.length] = trans;
    this.transitions = transitions2;

    WakeupCriterion c[] = new WakeupCriterion[transitions.length];
    for(int i=0; i<transitions.length; i++)
      c[i] = transitions[i].getWakeupCriterion();
    this.activeCriterion = new WakeupOr(c);

    if(debutThis)Logger.debug.println("DEBUG [State] addTransition() #criteria="+c.length+"");
   }

   /**
    * All tramsitions from here
    */
   private Transition transitions[];

   /**
    * An or of acc criteria of all transitions
    */
   private WakeupOr activeCriterion;

   public WakeupCondition getActiveCondition()
   {
    return(activeCriterion);
   }

   public State processStimulus(WakeupCriterion stimulus, Stack stack)
   {
    State next = this;
    for(int i=0; i<transitions.length; i++)
      if((next = transitions[i].checkAndExecute(stimulus, stack)) != null)
        break;


    return(next);
   }

   public State processStimulus( Enumeration criteria, Stack stack )
   {
    while(criteria.hasMoreElements())
      {
       WakeupCriterion c = (WakeupCriterion)criteria.nextElement();
       State s = processStimulus(c, stack);
       if(s != this)
         return(s);
      }
    return(this);
   }

  }

  /**
   * Class representing a possible
   * transition into another state
   */
  public static class Transition
  {
   public Transition(State nextState, WakeupCriterion crit, Action a)
   {
    this.next = nextState;
    this.crit = crit;
    this.action = ((a==null)?new NoopAction():a);
   }
   private State next;
   private WakeupCriterion crit;
   private Action action;

   public boolean canExecute(WakeupCriterion stimulus)
   {
    return(crit.hasTriggered());
   }

   public State execute(WakeupCriterion stimulus, Stack stack)
   {
    this.action.execute(stimulus, stack);
    return(next);
   }

   public WakeupCriterion getWakeupCriterion()
   {
    if(debutThis)Logger.debug.println("DEBUG [Transition] getWakeupCriterion()="+crit+"");
    return(crit);
   }

   /**
    * Check if we can execute and return either the
    * next state or null.
    */
   public State checkAndExecute(WakeupCriterion stimulus, Stack stack)
   {
    if(canExecute(stimulus))
      return(execute(stimulus, stack));

    return(null);
   }
  }


  /**
   * Actions are exwecuted uppon state-transitions
   */
  public interface Action
  {
   public void execute(WakeupCriterion stimulus, Stack stack);
  }

  public static class NoopAction implements Action
  {
   public void execute(WakeupCriterion stimulus, Stack stack)
    {};
  }

  public static class ListAction implements Action
  {
   private Action actions[];
   public ListAction()
   {
    actions = new Action[0];
   };
   public ListAction(Action actions[])
   {
    this.actions = actions;
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    for(int i=0; i<actions.length; i++)
       {
        actions[i].execute(stimulus, stack);
       }
   };
  }

  public static class DebugAction implements Action
  {
   String text;
   public DebugAction(String text)
   {
    this.text = text;
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    Logger.debug.println(text);
   };
  }

  /**
   * Invoke a method on an object to get a value for the stack.
   * (e.g. a position)
   */
  public static class PushValueAction implements Action
  {
   private Object obj;
   private Object args[];
   private Method method;

   /**
    * Just push the args onto the stack (one at a time)
    */
   public PushValueAction(Object args[])
   {
    this.obj = null;
    this.args = args;
    this.method = null;
   };


   public PushValueAction(Object obj, Object args[], Method method)
   {
    this.obj = obj;
    this.args = args;
    this.method = method;
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    try
      {
       if(obj == null)
        {
         if(args != null)
         for(int i=0; i<args.length; i++)
            stack.push(args[i]);
         return;
        }

       Object o = method.invoke(obj, args);
       if(o != null)
         stack.push(o);
      }
    catch(Exception e)
      {
       e.printStackTrace(/*Logger.err*/);
      }
   };
  }

  /**
   * Invoke a method on the stimulus to get a value for the stack.
   * (e.g. a position)
   */
  public static class StimulusPushValueAction implements Action
  {
   private Object args[];
   private Method method;

   public StimulusPushValueAction(Object args[], Method method)
   {
    this.args = args;
    this.method = method;
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    try
      {
       if(!method.getDeclaringClass().isInstance(stimulus))
         {
          Logger.err.println("ERROR: [StimulusPushValueAction] stimulus is not of a class compactible with method["+method.getDeclaringClass()+"] but of["+stimulus.getClass()+"]");
          return;
         }
       Object o = method.invoke(stimulus , args);
       if(o != null)
         stack.push(o);
      }
    catch(Exception e)
      {
       e.printStackTrace(/*Logger.err*/);
      }
   };
  }

  /**
   * Print any and all data about the stimulus of Logger.debug
   */
  public static class StimulusDebugAction implements Action
  {

   public StimulusDebugAction()
   {
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    try
      {
       Logger.debug.println("DEBUG: StimulusDebugAction was triggered with the stimulus \""+stimulus.toString()+"\"");
       java.lang.reflect.Field f[] = stimulus.getClass().getDeclaredFields();
       for(int i=0; i<f.length;i++)
          if(f[i].isAccessible())
            Logger.debug.println("DEBUG: stimulus."+f[i].getName()+"\""+f[i].get(stimulus)+"\"");

       java.beans.PropertyDescriptor pd[] = java.beans.Introspector.getBeanInfo(stimulus.getClass()).getPropertyDescriptors();
       for(int i=0; i<pd.length;i++)
           {
            java.lang.reflect.Method meth = pd[i].getReadMethod();
            if(meth != null)
              {
               Object value = null;
               try
                 {
                  value = meth.invoke(stimulus, new Object[0]);
                  if(value != null)
                    try
                      {
                       value = "["+value.getClass().toString()+"]"+value.toString();
                      }
                    catch(Exception e)
                      {
                       value = "["+value.getClass().toString()+"] Exception: "+e.toString();
                      }
                 }
               catch(Exception e)
                 {
                  value = "Exception: "+e.getMessage();
                 }
               Logger.debug.println("DEBUG: stimulus."+meth.getName()+"()=\""+(value==null?"null":value)+"\"");
              }
           }
      }
    catch(Exception e)
      {
       e.printStackTrace(/*Logger.err*/);
      }
   };
  }

  /**
   * Invoke a method on an object with calculated arguments.
   * (e.g. modify an object)
   */
  public static class InvokeAction implements Action
  { //TODO: use this gnu-expression-stuff to generate compiled expressions for the arguments using e.g. objects from the stack
   private Object obj;
//TODO   private Object args[];
   private Method method;

   public InvokeAction(Object obj, /*TODO Object args[],*/ Method method)
   {
    this.obj = obj;
//TODO    this.args = args;
    this.method = method;
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    try
      {
       Object o = method.invoke(obj, /*TODOargs*/new Object[0]);
       if(o != null)
         stack.push(o);
      }
    catch(Exception e)
      {
       e.printStackTrace(/*Logger.err*/);
      }
   };
  }

  /**
   * Invoke a method on an object with the first n Onjects on the stack
   * (e.g. modify an object)
   */
  public static class PopInvokeAction implements Action
  {
   private Object obj;
   private int popcount;
   private Method method;

   public PopInvokeAction(Object obj, int popcount, Method method)
   {
    this.obj = obj;
    this.popcount = popcount;
    this.method = method;
    if(!method.getDeclaringClass().isInstance(obj))
       Logger.err.println("ERROR: [PopInvokeAction] init() obj is not of a class compactible with method");
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    Object args[] = new Object[popcount];
    try
      {
       for(int i=0; i<args.length;i++)
         args[i] = stack.pop();
       Object o = method.invoke(obj, args);
       if(o != null)
         stack.push(o);
      }
    catch(Exception e)
      {
       try
         {
          Logger.err.println("ERROR: [PopInvokeAction] execute method="+(method==null?"<null>":method.getName())+" #args="+popcount+" "+(args.length>0?"arg0.class=\""+args[0].getClass().toString()+"\"":"")+" "+(method.getParameterTypes().length>0?" arg0.class should be \""+(method.getParameterTypes().length>0?method.getParameterTypes()[0].toString():"<0 parameters>")+"\"":""));
         }
       catch(Exception e2)
         { // ignore it, we have the stack-trace
         }
       e.printStackTrace(Logger.err);
      }
   };
  }


  /**
   * Invoke a method on the uppermost object of the stack
   * and pust it's return-value and maybe itself onto the stack
   * (e.g. modify an object) then push maybe the arguments on top of it
   */
  public static class StackInvokeAction implements Action
  {
   private boolean pushback;
   private boolean pushbackargs;
   private int popcount;
   private Method method;

   public StackInvokeAction(boolean pushback, boolean pushbackargs, int popcount, Method method)
   {
    this.pushback = pushback;
    this.pushbackargs = pushbackargs;
    this.popcount = popcount;
    this.method = method;
   };

   public void execute(WakeupCriterion stimulus, Stack stack)
   {
    try
      {
       Object obj = stack.pop();

      if(!method.getDeclaringClass().isInstance(obj))
         Logger.err.println("ERROR: [StackInvokeAction] obj is not of a class compactible with method["+method.getDeclaringClass()+"] but of["+obj.getClass()+"]");

       Object args[] = new Object[popcount];
       for(int i=0; i<args.length;i++)
         {
          args[i] = stack.pop();
          if(debutThis)Logger.err.println("DEBUG: [StackInvokeAction] poping up argument #"+i+" = ["+args[i].getClass()+"]\""+args[i]+"\"");
         }

       Object o = method.invoke(obj, args);
       if(o != null)
         stack.push(o);
       if(pushback)
         stack.push(obj);
       if(pushbackargs)
       for(int i=0; i<args.length;i++)
         {
          stack.push(args[i]);
          if(debutThis)Logger.err.println("DEBUG: [StackInvokeAction] pushing back argument #"+i+" = ["+args[i].getClass()+"]\""+args[i]+"\"");
         }
      }
    catch(Exception e)
      {
       e.printStackTrace(/*Logger.err*/);
      }
   };
  }





  //--------------------------------------------------------

  private Automata automata;


  public AutomataBehavior(Automata a)
  {
   this.automata = a;
  }

  public Automata getAutomata()
  {
   return(automata);
  }

  public void setAutomata(Automata automata)
  {
   this.automata = automata;
  }



  public void processStimulus(Enumeration param1)
  {
   if(debutThis)Logger.debug.println("DEBUG [AutomataBehavior] processStimulus()");
   automata.processStimulus(param1, this);
  }

  public void initialize()
  {
   if(debutThis)Logger.debug.println("DEBUG [AutomataBehavior] initialize()");
   automata.activate(this);
  }
}